// Copyright 2021 dfuse Platform Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cli

import (
	"fmt"
	"os"
	"strconv"
	"strings"
	"time"

	"github.com/streamingfast/dstore"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"github.com/streamingfast/bstream"
	"github.com/streamingfast/bstream/transform"
	dauthAuthenticator "github.com/streamingfast/dauth/authenticator"
	"github.com/streamingfast/dlauncher/launcher"
	"github.com/streamingfast/dmetering"
	"github.com/streamingfast/dmetrics"
	"github.com/streamingfast/firehose"
	firehoseApp "github.com/streamingfast/firehose/app/firehose"
	"github.com/streamingfast/logging"
	"github.com/streamingfast/sf-ethereum/filtering"
	sftransform "github.com/streamingfast/sf-ethereum/transform"
	"go.uber.org/zap"
)

var metricset = dmetrics.NewSet()
var headBlockNumMetric = metricset.NewHeadBlockNumber("firehose")
var headTimeDriftmetric = metricset.NewHeadTimeDrift("firehose")

func init() {
	appLogger := zap.NewNop()
	logging.Register("github.com/streamingfast/sf-ethereum/firehose", &appLogger)

	launcher.RegisterApp(&launcher.AppDef{
		ID:          "firehose",
		Title:       "Block Firehose",
		Description: "Provides on-demand filtered blocks, depends on common-blocks-store-url and common-blockstream-addr",
		MetricsID:   "merged-filter",
		Logger:      launcher.NewLoggingDef("github.com/streamingfast/sf-ethereum/firehose.*", nil),
		RegisterFlags: func(cmd *cobra.Command) error {
			cmd.Flags().String("firehose-grpc-listen-addr", FirehoseGRPCServingAddr, "Address on which the firehose will listen, appending * to the end of the listen address will start the server over an insecure TLS connection")
			cmd.Flags().StringSlice("firehose-blocks-store-urls", nil, "If non-empty, overrides common-blocks-store-url with a list of blocks stores")
			cmd.Flags().Duration("firehose-realtime-tolerance", 2*time.Minute, "Longest delay to consider this service as real-time (ready) on initialization")
			// irreversible indices
			cmd.Flags().String("firehose-irreversible-blocks-index-url", "", "If non-empty, will use this URL as a store to read irreversibility data on blocks and optimize replay")
			cmd.Flags().Bool("firehose-write-irreversible-blocks-index", false, "If set, will also create missing irreversibility indexes and write them to firehose-irreversible-blocks-index-url")
			cmd.Flags().IntSlice("firehose-irreversible-blocks-index-bundle-sizes", []int{100000, 10000, 1000, 100}, "list of sizes for irreversible block indices")
			// block indices
			cmd.Flags().String("firehose-block-index-url", "", "If non-empty, will use this URL as a store to load index data used by some transforms")
			cmd.Flags().IntSlice("firehose-block-index-sizes", []int{100000, 10000, 1000, 100}, "list of sizes for block indices")
			return nil
		},

		FactoryFunc: func(runtime *launcher.Runtime) (launcher.App, error) {
			sfDataDir := runtime.AbsDataDir
			tracker := runtime.Tracker.Clone()
			blockstreamAddr := viper.GetString("common-blockstream-addr")
			if blockstreamAddr != "" {
				tracker.AddGetter(bstream.BlockStreamLIBTarget, bstream.StreamLIBBlockRefGetter(blockstreamAddr))
			}

			// FIXME: That should be a shared dependencies across `Ethereum on StreamingFast`
			authenticator, err := dauthAuthenticator.New(viper.GetString("common-auth-plugin"))
			if err != nil {
				return nil, fmt.Errorf("unable to initialize dauth: %w", err)
			}

			// FIXME: That should be a shared dependencies across `Ethereum on StreamingFast`, it will avoid the need to call `dmetering.SetDefaultMeter`
			metering, err := dmetering.New(viper.GetString("common-metering-plugin"))
			if err != nil {
				return nil, fmt.Errorf("unable to initialize dmetering: %w", err)
			}
			dmetering.SetDefaultMeter(metering)

			firehoseBlocksStoreURLs := viper.GetStringSlice("firehose-blocks-store-urls")
			if len(firehoseBlocksStoreURLs) == 0 {
				firehoseBlocksStoreURLs = []string{MustReplaceDataDir(sfDataDir, viper.GetString("common-blocks-store-url"))}
			} else if len(firehoseBlocksStoreURLs) == 1 && strings.Contains(firehoseBlocksStoreURLs[0], ",") {
				// Providing multiple elements from config doesn't work with `viper.GetStringSlice`, so let's also handle the case where a single element has separator
				firehoseBlocksStoreURLs = strings.Split(firehoseBlocksStoreURLs[0], ",")
			}

			for i, url := range firehoseBlocksStoreURLs {
				firehoseBlocksStoreURLs[i] = MustReplaceDataDir(sfDataDir, url)
			}

			shutdownSignalDelay := viper.GetDuration("common-system-shutdown-signal-delay")
			grcpShutdownGracePeriod := time.Duration(0)
			if shutdownSignalDelay.Seconds() > 5 {
				grcpShutdownGracePeriod = shutdownSignalDelay - (5 * time.Second)
			}

			filterPreprocessorFactory := func(includeExpr, excludeExpr string) (bstream.PreprocessFunc, error) {
				filter, err := filtering.NewBlockFilter([]string{includeExpr}, []string{excludeExpr})
				if err != nil {
					return nil, fmt.Errorf("parsing filter expressions: %w", err)
				}

				preproc := &filtering.FilteringPreprocessor{Filter: filter}
				return preproc.PreprocessBlock, nil
			}

			if ll := os.Getenv("FIREHOSE_THREADS"); ll != "" {
				if llint, err := strconv.ParseInt(ll, 10, 32); err == nil {
					zlog.Info("setting blockstreamV2 parallel file downloads", zap.Int("ll", int(llint)))
					firehose.StreamBlocksParallelFiles = int(llint)
				}
			}

			indexStoreUrl := viper.GetString("firehose-block-index-url")
			var indexStore dstore.Store
			if indexStoreUrl != "" {
				s, err := dstore.NewStore(indexStoreUrl, "", "", false)
				if err != nil {
					return nil, fmt.Errorf("couldn't create indexStore: %w", err)
				}
				indexStore = s
			}

			var possibleIndexSizes []uint64
			for _, size := range viper.GetIntSlice("firehose-block-index-sizes") {
				if size < 0 {
					return nil, fmt.Errorf("invalid negative size for firehose-block-index-sizes: %d", size)
				}
				possibleIndexSizes = append(possibleIndexSizes, uint64(size))
			}

			registry := transform.NewRegistry()
			registry.Register(sftransform.BasicLogFilterFactory(indexStore, possibleIndexSizes))
			registry.Register(sftransform.LightBlockFilterFactory)

			var bundleSizes []uint64
			for _, size := range viper.GetIntSlice("firehose-irreversible-blocks-index-bundle-sizes") {
				if size < 0 {
					return nil, fmt.Errorf("invalid negative size for firehose-irreversible-blocks-index-bundle-sizes: %d", size)
				}
				bundleSizes = append(bundleSizes, uint64(size))
			}

			return firehoseApp.New(appLogger, &firehoseApp.Config{
				BlockStoreURLs:                  firehoseBlocksStoreURLs,
				BlockStreamAddr:                 blockstreamAddr,
				GRPCListenAddr:                  viper.GetString("firehose-grpc-listen-addr"),
				GRPCShutdownGracePeriod:         grcpShutdownGracePeriod,
				RealtimeTolerance:               viper.GetDuration("firehose-realtime-tolerance"),
				IrreversibleBlocksIndexStoreURL: viper.GetString("firehose-irreversible-blocks-index-url"),
				WriteIrreversibleBlocksIndex:    viper.GetBool("firehose-write-irreversible-blocks-index"),
				IrreversibleBlocksBundleSizes:   bundleSizes,
			}, &firehoseApp.Modules{
				Authenticator:             authenticator,
				FilterPreprocessorFactory: filterPreprocessorFactory,
				HeadTimeDriftMetric:       headTimeDriftmetric,
				HeadBlockNumberMetric:     headBlockNumMetric,
				Tracker:                   tracker,
				TransformRegistry:         registry,
			}), nil
		},
	})
}
